@@ -20,6 +20,7 @@
 #include <glib.h>
 
 #include <gnutls/gnutls.h>
+#include <gnutls/x509.h>
 
 #include "soup-ssl.h"
 #include "soup-misc.h"
@@ -78,6 +79,11 @@ verify_certificate (gnutls_session session, const char *hostname)
 		const gnutls_datum* cert_list;
 		int cert_list_size;
 		gnutls_x509_crt cert;
+
+		if (gnutls_x509_crt_init (&cert) < 0) {
+			g_warning ("Error initializing certificate.");
+			return FALSE;
+		}
       
 		cert_list = gnutls_certificate_get_peers (
 			session, &cert_list_size);
@@ -110,8 +116,14 @@ do_handshake (SoupGNUTLSChannel *chan, GError **err)
 	result = gnutls_handshake (chan->session);
 
 	if (result == GNUTLS_E_AGAIN ||
-	    result == GNUTLS_E_INTERRUPTED)
+	    result == GNUTLS_E_INTERRUPTED) {
+		g_set_error (err, SOUP_SSL_ERROR,
+			     (gnutls_record_get_direction (chan->session) ?
+			      SOUP_SSL_ERROR_HANDSHAKE_NEEDS_WRITE :
+			      SOUP_SSL_ERROR_HANDSHAKE_NEEDS_READ),
+			     "Handshaking...");
 		return G_IO_STATUS_AGAIN;
+	}
 
 	if (result < 0) {
 		g_set_error (err, G_IO_CHANNEL_ERROR,
@@ -242,9 +254,8 @@ soup_gnutls_close (GIOChannel  *channel,
 		int ret;
 
 		do {
-			ret = gnutls_bye (chan->session, GNUTLS_SHUT_RDWR);
-		} while (ret == GNUTLS_E_INTERRUPTED ||
-			 ret == GNUTLS_E_AGAIN);
+			ret = gnutls_bye (chan->session, GNUTLS_SHUT_WR);
+		} while (ret == GNUTLS_E_INTERRUPTED);
 	}
 
 	return chan->real_sock->funcs->io_close (channel, err);
